// Copyright 2020 TiKV Project Authors. Licensed under Apache-2.0.

use futures::future::Future;
use std::time::Duration;

use crate::pd_client::PdClient;
use raft::eraftpb::ConfChangeType;
use raftstore::store::QuorumAlgorithm;
use test_raftstore::*;

// Test QuorumAlgorithm::IntegrationOnHalfFail works as expected.
#[test]
fn test_integration_on_half_fail_quorum_fn() {
    let mut cluster = new_node_cluster(0, 5);
    cluster.cfg.raft_store.quorum_algorithm = QuorumAlgorithm::IntegrationOnHalfFail;
    cluster.run();
    cluster.must_put(b"k1", b"v0");

    // After peer 4 and 5 fail, no new leader could be elected.
    cluster.stop_node(4);
    cluster.stop_node(5);
    for _ in 0..500 {
        sleep_ms(10);
        if cluster.leader_of_region(1).is_none() {
            return;
        }
    }
    panic!("region 1 must lost leader because quorum fail");
}

#[test]
fn test_check_conf_change() {
    let mut cluster = new_node_cluster(0, 3);
    cluster.cfg.raft_store.quorum_algorithm = QuorumAlgorithm::IntegrationOnHalfFail;
    let rid = cluster.run_conf_change();

    cluster.pd_client.must_add_peer(rid, new_peer(2, 2));
    cluster.pd_client.must_add_peer(rid, new_learner_peer(3, 3));

    let r1 = cluster
        .pd_client
        .get_region_by_id(rid)
        .wait()
        .unwrap()
        .unwrap();

    // Stop peer 3 and compact logs on peer 1 and 2, so that peer 3 can't become voter.
    cluster.stop_node(3);
    for _ in 0..100 {
        cluster.must_put(b"k1", b"v1");
    }
    let compact_log = new_compact_log_request(100, 10);
    let req = new_admin_request(rid, r1.get_region_epoch(), compact_log);
    let res = cluster
        .call_command_on_leader(req, Duration::from_secs(3))
        .unwrap();
    assert!(!res.get_header().has_error());

    // Peer 3 can't be promoted.
    let cc = new_change_peer_request(ConfChangeType::AddNode, new_peer(3, 3));
    let req = new_admin_request(rid, r1.get_region_epoch(), cc);
    let res = cluster
        .call_command_on_leader(req, Duration::from_secs(3))
        .unwrap();
    assert!(res
        .get_header()
        .get_error()
        .get_message()
        .contains("unsafe to perform conf change peer"));
}
